/* $Id: tess_fist.c,v 1.8 1999/09/17 06:38:20 gareth Exp $ */

/*
 * Mesa 3-D graphics library
 * Version:  3.1
 *
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * $Log: tess_fist.c,v $
 * Revision 1.8  1999/09/17 06:38:20  gareth
 * Winding rule updates.
 *
 * Revision 1.7  1999/09/17 03:07:28  tjump
 * Win32 build req't updates
 *
 * Revision 1.6  1999/09/16 06:43:39  gareth
 * Misc winding rule bug fixes.
 *
 * Revision 1.5  1999/09/15 02:25:31  gareth
 * Added debugging output, fixed vector macro calls.
 *
 * Revision 1.4  1999/09/13 22:20:13  gareth
 * Fixed file headers.  Tracking down macro bugs.
 *
 */

/*****************************************************************************
 *
 * GLU 1.3 Polygon Tessellation - Implementation of FIST algorithm
 *
 * For more info on FIST, see:
 *  http://www.cosy.sbg.ac.at/~held/projects/triang/triang.html
 *
 * Gareth Hughes <garethh@lucent.com>, April 1999
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/glu.h>

#include "gluP.h"

#include "tess.h"
#include "tess_macros.h"
#include "tess_heap.h"
#include "tess_fist.h"
#if 0
#include "tess_preprocess.h"
#include "tess_grid.h"
#endif


/*****************************************************************************
 * Internal function prototypes:
 *****************************************************************************/

static GLenum compute_orientations( GLUtesselator *tobj );
static GLenum sort_vertices( GLUtesselator *tobj );
static GLenum transform_build_bridges( GLUtesselator *tobj );
static GLenum classify_angles( GLUtesselator *tobj );
static GLenum tessellate_contours( GLUtesselator *tobj );
static GLenum output_contours( GLUtesselator *tobj );

static GLdouble	point_line_test( GLdouble u[2], GLdouble v[2], GLdouble p[2] );
static GLboolean point_triangle_test( tess_vertex_t *triangle,
				      tess_vertex_t *point );

GLdouble angle_3dv( GLdouble va[2], GLdouble vb[2], GLdouble vc[2] );

static void tess_begin_callback( GLUtesselator *tobj, GLenum mode,
				 void *data );
static void tess_vertex_callback( GLUtesselator *tobj, void *vertex_data,
				  void *polygon_data );
static void tess_end_callback( GLUtesselator *tobj, void *data );

static void cleanup( GLUtesselator *tobj );

static void contour_dump( tess_contour_t *contour );


/*****************************************************************************
 * Internal type definitions:
 *****************************************************************************/

typedef struct
{
    tess_vertex_t	*vertex;
    GLdouble		dist;
} fist_vecdist_t;

#define RV_INSERT(t, r)			\
	hashtable_insert( (t), ((int) r), ((void *) r) )
#define RV_DELETE(t, r)			\
	hashtable_delete( (t), ((int) r) )



/*****************************************************************************
 *
 *		    FAST INDUSTRIAL-STRENGTH TRIANGULATION (FIST)
 *
 *****************************************************************************/


/*****************************************************************************
 * fist_tessellation
 *
 * The main algorithm.  This is the core triangulation code that implements
 * the Fast Industrial-Strength Triangulation algorithm.
 *****************************************************************************/
GLenum fist_tessellation( GLUtesselator *tobj )
{
    DEBUGP( 0, ( "  -> fist_tessellation( tobj:%p )\n", tobj ) );

#if 0
    /* Calculate and add any intersection points. */
    tess_calc_intersections( tobj );
#endif

    /* Sort and renumber the vertices of all the polygons. */
    sort_vertices( tobj );

    /* Compute and adjust the orientations of all the polygons. */
    compute_orientations( tobj );

#if 0
    /* Do those fancy calculations for the GLU winding rules. */
    tess_preprocess_contours( tobj );
#endif

    /*
     * Transform multiple polygons into a single polygon by building all
     *  bridges.
     */
    if ( tobj->contour_count > 1 )
    {
	contour_dump( tobj->contours );

	transform_build_bridges( tobj );
    }

    /* Classify all of the angles of the polygon. */
    classify_angles( tobj );

    /* Output the damned thing! */
    output_contours( tobj );

    /* Tessellate all the contours that make up the polygon. */
    tessellate_contours( tobj );

    /* Cleanup after tessellation is complete. */
    cleanup( tobj );

    DEBUGP( 0, ( "    <- fist_tessellation( tobj:%p )\n", tobj ) );
    return tobj->error;
}


/*****************************************************************************
 * fist_recovery_process
 *
 * The main recovery algorithm.  This is the multi-phase recovery process the
 * FIST algorithm uses when things start going wrong.  It ensures a
 * topologically-valid triangulation is produced, no matter what.
 *****************************************************************************/
GLenum fist_recovery_process( GLUtesselator *tobj )
{
    /* FIXME: Implement the recovery process - GH */
    DEBUGP( 0, ( "*** error 8 ***\n" ) );
    tess_error_callback( tobj, GLU_TESS_ERROR8, NULL );

    return GLU_ERROR;
}



/*****************************************************************************
 *
 *				INTERNAL FUNCTIONS
 *
 *****************************************************************************/


/*****************************************************************************
 * compute_orientations
 *
 * The orientations of each contour loop previously calculated are inspected
 * and adjusted so that the final triangulation will be CCW with respect to
 * the normal of the triangulation plane.
 *****************************************************************************/
static int compare_contour_areas( const void *c1, const void *c2 );

static GLenum compute_orientations( GLUtesselator *tobj )
{
    tess_contour_t	**sorted_contours;
    tess_contour_t	*current;
    GLuint		i;

    DEBUGP( 2, ( "    -> compute_orientations( tobj:%p )\n", tobj ) );

    if ( tobj->contour_count > 1 )
    {
	/*
	 * If we have multiple contours, sort them by their size to determine
	 *  which one is the exterior.  Make sure this exterior contour is
	 *  oriented CCW with respect to the tessellation normal, and if it
	 *  is not oriented this way reverse all the contours.
	 *
	 * FIXME: Is this the best place to do this?  I think the winding
	 *  rule stuff will work out the exteriour contour better than this.
	 */

	if ( ( sorted_contours = (tess_contour_t **)
	       malloc( tobj->contour_count * sizeof(tess_contour_t *) ) )
	     == NULL )
	{
	    DEBUGP( 0, ( "*** memory error ***\n" ) );
	    tess_error_callback( tobj, GLU_OUT_OF_MEMORY, NULL );
	    return GLU_ERROR;
	}

	current = tobj->contours;
	i = 0;

	while ( i < tobj->contour_count )
	{
	    sorted_contours[ i++ ] = current;
	    current = current->next;
	}

	/* Sort the contours by their area. */
	qsort( sorted_contours, tobj->contour_count,
	       sizeof(tess_contour_t *), compare_contour_areas );

	tobj->contours     = sorted_contours[ 0 ];
	tobj->last_contour = sorted_contours[ tobj->contour_count - 1 ];

	current = tobj->contours;

	for ( i = 1; i < tobj->contour_count; i++ )
	{
	    current->next = sorted_contours[ i ];
	    sorted_contours[ i ]->previous = current;

	    current = current->next;
	}

	/* Wrap the contour list. */
	tobj->last_contour->next = tobj->contours;
	tobj->contours->previous = tobj->last_contour;

	if ( sorted_contours ) {
	    free( sorted_contours );
	}
    }

    DEBUGP( 2, ( "    <- compute_orientations( tobj:%p )\n", tobj ) );
    return GLU_NO_ERROR;
}

/*****************************************************************************
 * compare_contour_areas
 *****************************************************************************/
static int compare_contour_areas( const void *c1, const void *c2 )
{
    GLdouble	area1, area2;

    area1 = (*((tess_contour_t **) c1))->area;
    area2 = (*((tess_contour_t **) c2))->area;

    if ( area1 < area2 ) {
	return 1;
    }
    if ( area1 > area2 ) {
	return -1;
    }
    return 0;
}


/*****************************************************************************
 * sort_vertices
 *
 * Sort all the vertices from all contours lexicographically, first by
 * x-coordinate, and then by y-coordinate.  Duplicates are removed from this
 * array, so that coincident vertices have the same index in the array.
 *****************************************************************************/
static int compare_vertices( const void *v1, const void *v2 );

static GLenum sort_vertices( GLUtesselator *tobj )
{
    tess_contour_t	*current;
    tess_vertex_t	*vertex;
    GLuint		i, j, n = 0, vertex_count, removed;

    DEBUGP( 2, ( "    -> sort_vertices( tobj:%p )\n", tobj ) );

    /* Allocate the sorted vertices array. */

    if ( ( tobj->sorted_vertices = (tess_vertex_t **)
	  malloc( tobj->vertex_count * sizeof(tess_vertex_t *) ) ) == NULL )
    {
	DEBUGP( 0, ( "*** memory error ***\n" ) );
	tess_error_callback( tobj, GLU_OUT_OF_MEMORY, NULL );
	return GLU_ERROR;
    }

    /* Add each vertex from each contour to the array for sorting. */

    for ( current = tobj->contours, i = 0;
	  i < tobj->contour_count; current = current->next, i++ )
    {
	for ( vertex = current->vertices, j = 0;
	      j < current->vertex_count; vertex = vertex->next, j++ )
	{
	    tobj->sorted_vertices[ n++ ] = vertex;
	}
    }

    /* Sort the array of vertices. */
    qsort( tobj->sorted_vertices, tobj->vertex_count,
	   sizeof(tess_vertex_t *), compare_vertices );

    tobj->sorted_vertices[0]->index = 0;

    vertex_count = tobj->vertex_count;
    removed = 0;
    i = 1;

    /* Scan through the array, removing all duplicate entries, and assign
     *  each vertex its index in the array. */

    while ( i < vertex_count )
    {
	tobj->sorted_vertices[ i ] = tobj->sorted_vertices[ i + removed ];

	if ( COMPARE_3DV( tobj->sorted_vertices[ i ]->coords,
			  tobj->sorted_vertices[ i - 1 ]->coords ) )
	{
	    tobj->sorted_vertices[ i ]->index = i - 1;

	    removed++;
	    vertex_count--;
	}
	else
	{
	    tobj->sorted_vertices[ i ]->index = i;
	    i++;
	}
    }

    /*
     * Shuffle each contour around so the head of the vertex list is the
     *  leftmost vertex.
     */
    current = tobj->contours;

    for ( i = 0; i < tobj->contour_count; i++ )
    {
	tess_vertex_t	*left = current->vertices;
	tess_vertex_t	*vertex = current->vertices->next;
	GLuint		j;

	for ( j = 1; j < current->vertex_count; j++ )
	{
	    if ( vertex->index < left->index )
	    {
		left = vertex;
	    }
	    vertex = vertex->next;
	}
	current->vertices = left;
	current->last_vertex = left->previous;

	current = current->next;
    }

    DEBUGP( 2, ( "    <- sort_vertices( tobj:%p )\n", tobj ) );
    return GLU_NO_ERROR;
}

/*****************************************************************************
 * compare_vertices
 *****************************************************************************/
static int compare_vertices( const void *v1, const void *v2 )
{
    tess_vertex_t	*vertex1 = *((tess_vertex_t **) v1);
    tess_vertex_t	*vertex2 = *((tess_vertex_t **) v2);
    int			ret = 0;

    ret = vertex1->v[X] - vertex2->v[X];

    if ( ret == 0 ) {
	ret = vertex1->v[Y] - vertex2->v[Y];
    }
    return ret;
}


/*****************************************************************************
 * transform_build_bridges
 *
 * Multiply-connected polygonal areas are transformed into a single degenerate
 * contour by iteratively linking the internal island loops with the outer
 * boundary by means of contour "bridges".  Thus, the triangulation need only
 * deal with a single contour when this process has been completed.
 *****************************************************************************/
static int compare_contour_left_vertices( const void *c1, const void *c2 );
static int compare_vertex_distances( const void *v1, const void *v2 );

static GLenum transform_build_bridges( GLUtesselator *tobj )
{
    tess_contour_t	**sorted;
    tess_contour_t	*contour = tobj->contours;
    GLuint		i, num_sorted;

    DEBUGP( 2, ( "    -> transform_build_bridges( tobj:%p )\n", tobj ) );

    if ( ( sorted = (tess_contour_t **)
	   malloc( tobj->contour_count * sizeof(tess_contour_t *) ) ) == NULL )
    {
	DEBUGP( 0, ( "*** memory error ***\n" ) );
	tess_error_callback( tobj, GLU_OUT_OF_MEMORY, NULL );
	return GLU_ERROR;
    }

    for ( contour = tobj->contours, num_sorted = 0, i = 0;
	  i < tobj->contour_count; contour = contour->next, i++ )
    {
	if ( contour->orientation == GLU_CW )
	{
	    sorted[ num_sorted++ ] = contour;
	}
    }

    /* Sort the contours by their leftmost vertices. */
    qsort( sorted, num_sorted, sizeof(tess_contour_t *),
	   compare_contour_left_vertices );

    /* Reset contour pointer.  This is the contour we want to add
        the islands to. */
    contour = tobj->contours;

    for ( i = 0; i < num_sorted; i++ )
    {
	fist_vecdist_t	*closest;
	tess_vertex_t	*vertex, *left_vertex = sorted[ i ]->vertices;
	tess_vertex_t	*new_edge[2];
	GLuint		j, num_closest = left_vertex->index;

	if ( ( closest = (fist_vecdist_t *)
	       malloc( num_closest * sizeof(fist_vecdist_t) ) ) == NULL )
	{
	    DEBUGP( 0, ( "*** memory error ***\n" ) );
	    tess_error_callback( tobj, GLU_OUT_OF_MEMORY, NULL );
	    return GLU_ERROR;
	}

	for ( vertex = contour->vertices, j = 0;
	      j < contour->vertex_count; vertex = vertex->next, j++ )
	{
	    if ( vertex->index < left_vertex->index )
	    {
		closest[ vertex->index ].vertex = vertex;
		closest[ vertex->index ].dist =
		    LEN_SCALAR( vertex->v[X] - left_vertex->v[X],
				vertex->v[Y] - left_vertex->v[Y] );
	    }
	}

	qsort( closest, num_closest, sizeof(fist_vecdist_t),
	       compare_vertex_distances );

	/*
	 * FIXME:  Check validity of added edge...
	 */

	new_edge[0] = (tess_vertex_t *) malloc( sizeof(tess_vertex_t) );
	new_edge[1] = (tess_vertex_t *) malloc( sizeof(tess_vertex_t) );

	if ( ( new_edge[0] == NULL ) || ( new_edge[1] == NULL ) )
	{
	    DEBUGP( 0, ( "*** memory error ***\n" ) );
	    tess_error_callback( tobj, GLU_OUT_OF_MEMORY, NULL );
	    return GLU_ERROR;
	}

	/* FIXME: Describe what the hell is going on in here. */

	memcpy( new_edge[0], left_vertex, sizeof(tess_vertex_t) );
	memcpy( new_edge[1], closest[0].vertex, sizeof(tess_vertex_t) );

	new_edge[0]->next = new_edge[1];
	new_edge[1]->previous = new_edge[0];

	left_vertex->previous->next = new_edge[0];

	closest[0].vertex->next = left_vertex;
	left_vertex->previous = closest[0].vertex;

	contour->vertex_count += 2 + sorted[ i ]->vertex_count;

	sorted[ i ]->previous->next = sorted[ i ]->next;
	sorted[ i ]->next->previous = sorted[ i ]->previous;

	if ( sorted[ i ] == tobj->last_contour )
	{
	    tobj->last_contour = sorted[ i ]->previous;
	}

	tobj->contour_count--;

	if ( closest ) {
	    free( closest );
	    closest = NULL;
	}
	if ( sorted[ i ] ) {
	    free( sorted[ i ] );
	    sorted[ i ] = NULL;
	}
    }

    if ( sorted ) {
	free( sorted );
    }

    DEBUGP( 2, ( "    <- transform_build_bridges( tobj:%p )\n", tobj ) );
    return GLU_NO_ERROR;
}

/*****************************************************************************
 * compare_contour_left_vertices
 *****************************************************************************/
static int compare_contour_left_vertices( const void *c1, const void *c2 )
{
    tess_vertex_t	*vertex1 = (*((tess_contour_t **) c1))->vertices;
    tess_vertex_t	*vertex2 = (*((tess_contour_t **) c2))->vertices;

    return vertex1->index - vertex2->index;
}

/*****************************************************************************
 * compare_vertex_distances
 *****************************************************************************/
static int compare_vertex_distances( const void *v1, const void *v2 )
{
    fist_vecdist_t	*vecdist1 = (fist_vecdist_t *) v1;
    fist_vecdist_t	*vecdist2 = (fist_vecdist_t *) v2;

    if ( ( vecdist2->dist - vecdist1->dist ) < EQUAL_EPSILON ) {
	return 1;
    } else {
	return -1;
    }
}


/*****************************************************************************
 * classify_angles
 *
 * All the internal angles of the contours are classified as convex or reflex.
 * This enables the geometric hashing that will significantly speed up the
 * triangulation process.
 *****************************************************************************/
static void classify_vertex( tess_contour_t *contour, tess_vertex_t *vertex );

static GLenum classify_angles( GLUtesselator *tobj )
{
    tess_contour_t	*contour;
    tess_vertex_t	*vertex;
    GLuint		i, j;

    DEBUGP( 2, ( "    -> classify_angles( tobj:%p )\n", tobj ) );

    for ( contour = tobj->contours, i = 0;
	  i < tobj->contour_count; contour = contour->next, i++ )
    {
	if ( ! contour->reflex_vertices ) {
	    contour->reflex_vertices = hashtable_init( HT_DEFAULT_SIZE );
	}

	for ( vertex = contour->vertices, j = 0;
	      j < contour->vertex_count; vertex = vertex->next, j++ )
	{
	    /*
	     * We need to break the classification out of the loops so
	     *  we can re-classify the vertices that change when an ear
	     *  is clipped.
	     */
	    classify_vertex( contour, vertex );
	}
    }

    DEBUGP( 2, ( "    <- classify_angles( tobj:%p )\n", tobj ) );
    return GLU_NO_ERROR;
}

/*****************************************************************************
 * classify_vertex
 *****************************************************************************/
static void classify_vertex( tess_contour_t *contour, tess_vertex_t *vertex )
{
    vertex->angle = point_line_test( vertex->previous->v,
				     vertex->v, vertex->next->v );

    DEBUGP( 0, ( "         classifying v: %d angle: %.2f\n", vertex->index, vertex->angle ) );

    /*
     * We have three cases here - note that 'angle' isn't really
     *  the angle between the edges meeting at the vertex, just
     *  an indication of the shape using the sign.
     *
     *  1) vertex->angle > 0  means angle < PI
     *  2) vertex->angle == 0 means angle == PI
     *  3) vertex->angle < 0  means angle > PI, or reflex
     *
     * So, what we want to grab here are all the vertices with
     *  an 'angle' that is less than zero.  We use our epsilon
     *  value to allow for some rounding errors in the calculation
     * FIXME: Is this epsilon value small enough?
     */
    if ( vertex->angle < -EQUAL_EPSILON ) {
 	RV_INSERT( contour->reflex_vertices, vertex );
    }
}


/*****************************************************************************
 * tessellate_contours
 *
 * Actually perform the ear clipping on the contours.
 *****************************************************************************/
static GLenum determine_ears( GLUtesselator *tobj, tess_contour_t *contour );
static GLboolean earity_test( tess_contour_t *contour, tess_vertex_t *vertex );
static GLdouble shape_classifier( tess_vertex_t *vertex );
static heap_elt_t *add_ear_to_heap( heap_t *heap, tess_vertex_t *vertex );

static GLenum tessellate_contours( GLUtesselator *tobj )
{
    tess_contour_t	*contour = tobj->contours;

    DEBUGP( 0, ( "    -> tessellate_contours( tobj:%p )\n", tobj ) );

    /*
     * Allocate the priority queue for the ears of the polgyon.  We use
     *  a fairly standard implementation of a heap for this queue, which
     *  is sorted by how close an ear is to an equilateral triangle.
     *  This will arguably give a higher-quality tessellation, at the cost
     *  of slightly reduced preformance.
     */
    tobj->ears = heap_init();

    /* Break the contour loop into a standard doubly-linked list. */
    tobj->last_contour->next = NULL;
    tobj->contours->previous = NULL;

    /*
     * Start outputting the triangles.  At the moment, we just ouput the
     *  ears when we get them.  Could we ouput them in strip order?  It
     *  would be nice, but it really makes the implementation of the
     *  algorithm less general than it is now.  Is the implementation
     *  too general now?  Quite possibly...
     */
    tess_begin_callback( tobj, GL_TRIANGLES, NULL );

    while ( contour )
    {
	GLboolean	clipped = GL_FALSE;

	determine_ears( tobj, contour );

	while ( contour->vertex_count > 3 )
	{
	    if ( tobj->ears->size > 0 )
	    {
		heap_elt_t	*element = heap_extract_max( tobj->ears );
		heap_elt_t	*next = NULL, *previous = NULL;
		tess_vertex_t	*vertex = HEAP_VERTEX( element->ptr );

		/* Remove ear from current contour... */
		DEBUGP(0, ("           nv: %2d hs: %d ear: (%d, %d, %d)\n", contour->vertex_count, tobj->ears->size, vertex->previous->index, vertex->index, vertex->next->index));

		clipped = GL_TRUE;

		/* Output the triangle. */
		tess_vertex_callback( tobj, vertex->coords, vertex->data );
		tess_vertex_callback( tobj, vertex->next->coords,
				      vertex->next->data );
		tess_vertex_callback( tobj, vertex->previous->coords,
				      vertex->previous->data );

		/* Remove the vertex from the contour as required. */
		vertex->previous->next = vertex->next;
		vertex->next->previous = vertex->previous;

		if ( contour->vertices == vertex ) {
		    contour->vertices = vertex->next;
		}
		contour->vertex_count--;

		/* Update the ear heap for the previous vertex. */
		if ( element->previous )
		{
		    /*
		     * An existing entry based on the previous vertex will
		     *  no longer be valid, so we need to remove it now.
		     */
		    previous = heap_delete( tobj->ears,
					    element->previous->index );

		    if ( ! previous ) continue;

		    /*
		     * We can now determine if clipping the current ear
		     *  will form a new ear based on the previous vertex.
		     *  If such a new ear is formed, add it to the ear
		     *  queue.  Otherwise, just clean up.
		     */
		    if ( earity_test( contour, HEAP_VERTEX( previous->ptr ) ) )
		    {
			previous->value =
			    shape_classifier( HEAP_VERTEX( previous->ptr ) );

			heap_insert( tobj->ears, previous );
		    }
		    else
		    {
			if ( previous->previous ) {
			    previous->previous->next = NULL;
			}
			free( previous );
			previous = NULL;
		    }
		}

		/* Update the ear heap for the next vertex. */
		if ( element->next )
		{
		    /*
		     * An existing entry based on the next vertex will
		     *  no longer be valid, so we need to remove it now.
		     */
		    next = heap_delete( tobj->ears, element->next->index );

		    if ( ! next ) continue;

		    /*
		     * We can now determine if clipping the current ear
		     *  will form a new ear based on the next vertex.  If
		     *  such a new ear is formed, add it to the ear queue.
		     *  Otherwise, just clean up.
		     */
		    if ( earity_test( contour, HEAP_VERTEX( next->ptr ) ) )
		    {
			next->value =
			    shape_classifier( HEAP_VERTEX( next->ptr ) );

			heap_insert( tobj->ears, next );
		    }
		    else
		    {
			if ( next->next ) {
			    next->next->previous = NULL;
			}
			free( next );
			next = NULL;
		    }
		}

		/* Update the previous and next ear queue entries. */
		if ( previous && next )
		{
		    previous->next = next;
		    next->previous = previous;
		}
		else if ( previous )
		{
		    previous->next = NULL;
		}
		else if ( next )
		{
		    next->previous = NULL;
		}

		/*
		 * Update the reflex vertex set.  We first need to remove
		 *  any entries for the three vertices we have touched,
		 *  and then re-classify the two vertices that remain in
		 *  the contour.
		 */
		RV_DELETE( contour->reflex_vertices, vertex->previous );
		RV_DELETE( contour->reflex_vertices, vertex );
		RV_DELETE( contour->reflex_vertices, vertex->next );

		classify_vertex( contour, vertex->previous );
		classify_vertex( contour, vertex->next );

		free( vertex );
		free( element );
	    }
	    else if ( clipped )
	    {
		clipped = GL_FALSE;
		determine_ears( tobj, contour );
	    }
	    else
	    {
		fist_recovery_process( tobj );
	    }
	}

	if ( contour->vertex_count == 3 )
	{
	    tess_vertex_t	*vertex = contour->vertices;

	    /* Output the last triangle. */
	    tess_vertex_callback( tobj, vertex->data, NULL );
	    tess_vertex_callback( tobj, vertex->next->data, NULL );
	    tess_vertex_callback( tobj, vertex->previous->data, NULL );

	    free( vertex->previous );
	    free( vertex->next );
	    free( vertex );
	}

	/* Move on to the next contour. */
	contour->vertices = NULL;
	contour->last_vertex = NULL;

	contour = contour->next;
    }

    /* Finish outputting the triangles. */
    tess_end_callback( tobj, NULL );

    DEBUGP( 0, ( "    <- tessellate_contours( tobj:%p )\n", tobj ) );
    return GLU_NO_ERROR;
}

/*****************************************************************************
 * determine_ears
 *
 * Determine the potential ears for the given contour.
 *****************************************************************************/
static GLenum determine_ears( GLUtesselator *tobj, tess_contour_t *contour )
{
    tess_vertex_t	*vertex = contour->vertices;
    heap_elt_t		*element = NULL, *previous = NULL, *first = NULL;
    GLuint		i;

    DEBUGP( 0, ( "      -> determine_ears( tobj:%p )\n", tobj ) );

    contour_dump( contour );

    for ( i = 0; i < contour->vertex_count; i++ )
    {
	if ( earity_test( contour, vertex ) )
	{
	    element = add_ear_to_heap( tobj->ears, vertex );

	    if ( element == NULL )
	    {
		DEBUGP( 0, ( "*** memory error ***\n" ) );
		tess_error_callback( tobj, GLU_OUT_OF_MEMORY, NULL );
		return GLU_ERROR;
	    }

	    element->previous = previous;

	    if ( previous ) {
		previous->next = element;
	    }

	    previous = element;

	    if ( i == 0 ) {
		first = element;
	    }
	}
	else
	{
	    previous = NULL;
	}

	vertex = vertex->next;
    }

    if ( previous ) {
	previous->next = first;
    }

    DEBUGP( 0, ( "      <- determine_ears( tobj:%p )\n", tobj ) );
    return GLU_NO_ERROR;
}

/*****************************************************************************
 * earity_test
 *
 * Is the given vertex part of an ear?  Test all the reflex vertices of the
 *  contour against the ear, and if none are inside the ear we have a winner.
 *****************************************************************************/
static GLboolean earity_test( tess_contour_t *contour, tess_vertex_t *vertex )
{
    hashtable_t	*table = contour->reflex_vertices;
    GLuint	i;

    if ( ( vertex->angle < EQUAL_EPSILON ) || ( ! table ) )
    {
	return GL_FALSE;
    }
    else
    {
	for ( i = 0; i < table->size; i++ )
	{
	    hashtable_elt_t	*elt = table->elements[ i ];

	    while ( elt )
	    {
		tess_vertex_t	*reflex = (tess_vertex_t *) elt->ptr;

		if ( point_triangle_test( vertex, reflex ) ) {
		    return GL_FALSE;
		}

		elt = elt->next;
	    }
	}

	return GL_TRUE;
    }
}

/*****************************************************************************
 * shape_classifier
 *
 * Return the minimum angle of the triangle made up by vertex, vertex->next
 *  and vertex->previous.
 *****************************************************************************/
static GLdouble shape_classifier( tess_vertex_t *vertex )
{
    GLdouble	min, current;

    min = angle_3dv( vertex->previous->v, vertex->v, vertex->next->v );

    current = angle_3dv( vertex->v, vertex->next->v, vertex->previous->v );
    if ( current < min ) { min = current; }

    current = angle_3dv( vertex->next->v, vertex->previous->v, vertex->v );
    if ( current < min ) { min = current; }

    return min;
}


/*****************************************************************************
 * add_ear_to_heap
 *
 * Add the ear containing the given vertex to the heap.
 *****************************************************************************/
static heap_elt_t *add_ear_to_heap( heap_t *heap, tess_vertex_t *vertex )
{
    heap_elt_t	*element;

    if ( (element = (heap_elt_t *) malloc( sizeof(heap_elt_t) ) ) == NULL ) {
	return NULL;
    }

    element->value = shape_classifier( vertex );
    element->ptr = (void *) vertex;
    element->index = 0;

    element->next = NULL;
    element->previous = NULL;

    heap_insert( heap, element );

    return element;
}



/*****************************************************************************
 * output_contours
 *
 * Output the tessellated contours.  Don't think we really need this anymore.
 *****************************************************************************/
static GLenum output_contours( GLUtesselator *tobj )
{
    tess_contour_t	*contour;
    tess_vertex_t	*vertex;
    GLuint		i, j;

    for ( contour = tobj->contours, i = 0;
	  i < tobj->contour_count; contour = contour->next, i++ )
    {
	tess_begin_callback( tobj, GL_LINE_LOOP, NULL );

	for ( vertex = contour->vertices, j = 0;
	      j < contour->vertex_count; vertex = vertex->next, j++ )
	{
	    tess_vertex_callback( tobj, vertex->coords, vertex->data );
	}
	tess_end_callback( tobj, NULL );
    }

    return GLU_NO_ERROR;
}



/*****************************************************************************
 * point_line_test
 *
 * Return positive if p is on the left of the line segment u -> v, negative
 * if p is on the right of u -> v, and 0 if p is on u -> v.
 *****************************************************************************/
static GLdouble point_line_test( GLdouble u[2], GLdouble v[2], GLdouble p[2] )
{
    return ( ( u[Y] - v[Y] ) * p[X] +
	     ( v[X] - u[X] ) * p[Y] +
	     ( u[X] * v[Y] - u[Y] * v[X] ) );
}

/*****************************************************************************
 * point_triangle_test
 *
 * Determine if the given point is inside the triangle given by the points
 * triangle, triangle->next and triangle->previous.
 *****************************************************************************/
static GLboolean point_triangle_test( tess_vertex_t *triangle,
				      tess_vertex_t *point )
{
    if ( ( point_line_test( triangle->v, triangle->next->v,
			    point->v ) > EQUAL_EPSILON ) &&
	 ( point_line_test( triangle->next->v, triangle->previous->v,
			    point->v ) > EQUAL_EPSILON ) &&
	 ( point_line_test( triangle->previous->v, triangle->v,
			    point->v ) > EQUAL_EPSILON ) )
    {
	return GL_TRUE;
    } else {
	return GL_FALSE;
    }
}

/*****************************************************************************
 * angle_3dv
 *
 * Calculate the interior angle between the line segments va->vb and vb->vc.
 *****************************************************************************/
GLdouble angle_3dv( GLdouble va[2], GLdouble vb[2], GLdouble vc[2] )
{
    GLdouble	u[2], v[2];
    GLdouble	ret;

    u[X] = va[X] - vb[X]; u[Y] = va[Y] - vb[Y];
    v[X] = vc[X] - vb[X]; v[Y] = vc[Y] - vb[Y] ;

    ret = (GLdouble) ( u[X] * v[X] + u[Y] * v[Y] ) /
	( LEN_SCALAR( u[X], u[Y] ) * LEN_SCALAR( v[X], v[Y] ) );

    return (GLdouble) acos( ret ) * 180 / M_PI;
}


/*****************************************************************************
 *
 *			TESSELLATION CALLBACKS
 *
 *****************************************************************************/

static void tess_begin_callback( GLUtesselator *tobj, GLenum mode, void *data )
{
    if ( tobj->callbacks.beginData != NULL )
    {
	( tobj->callbacks.beginData )( mode, data );
    }
    else if ( tobj->callbacks.begin != NULL )
    {
	( tobj->callbacks.begin )( mode );
    }
}

static void tess_vertex_callback( GLUtesselator *tobj, void *vertex_data,
				  void *polygon_data )
{
    if ( tobj->callbacks.vertexData != NULL )
    {
	( tobj->callbacks.vertexData )( vertex_data, polygon_data );
    }
    else if ( tobj->callbacks.vertex != NULL )
    {
	( tobj->callbacks.vertex )( vertex_data );
    }
}

static void tess_end_callback( GLUtesselator *tobj, void *data )
{
    if ( tobj->callbacks.endData != NULL )
    {
	( tobj->callbacks.endData )( data );
    }
    else if ( tobj->callbacks.end != NULL )
    {
	( tobj->callbacks.end )();
    }
}



/*****************************************************************************
 * cleanup
 *
 * Clean up after the triangulation has taken place.  All memory allocated
 * by the triangulation should be freed before control is handed back to the
 * GLU functions.
 *****************************************************************************/
static void cleanup( GLUtesselator *tobj )
{
    tess_contour_t	*contour = tobj->contours;
    GLuint		i;

    DEBUGP( 1, ( "    -> cleanup( tobj:%p )\n", tobj ) );

    if ( tobj->sorted_vertices != NULL ) {
	free( tobj->sorted_vertices );
    }
    tobj->sorted_vertices = NULL;

    heap_cleanup( tobj->ears );
    tobj->ears = NULL;

    for ( i = 0; i < tobj->contour_count; i++ )
    {
	hashtable_cleanup( contour->reflex_vertices );

	contour = contour->next;
    }

    DEBUGP( 1, ( "    <- cleanup( tobj:%p )\n", tobj ) );
}



/*****************************************************************************
 * contour_dump
 *
 * Dump all the vertices, reflex vertices etc for the given contour.
 *****************************************************************************/
static void contour_dump( tess_contour_t *contour )
{
    tess_vertex_t *vertex = contour->vertices;
#ifdef DEBUG
    GLuint i;

    fprintf( stderr, "\ncontour %p - nv: %d\n", contour, contour->vertex_count );

    for ( i = 0; i < contour->vertex_count; i++ )
    {
	fprintf( stderr, "  v: %d (%.2f, %.2f) (%d, %d, %d) label: %d\n", vertex->index, vertex->v[X], vertex->v[Y], vertex->previous->index, vertex->index, vertex->next->index, vertex->label );
	vertex = vertex->next;
    }

    fprintf( stderr, "\n" );

    if ( contour->reflex_vertices )
    {
	hashtable_t	*table = contour->reflex_vertices;

	if ( table->nelements > 0 ) {
	    fprintf( stderr, "  r:" );
	}
	for ( i = 0; i < table->size; i++ )
	{
	    hashtable_elt_t	*elt = table->elements[ i ];
	    
	    while ( elt )
	    {
		tess_vertex_t	*v = HASH_VERTEX( elt->ptr );
		fprintf( stderr, " %d", v->index );
	
		elt = elt->next;
	    }
	}
	if ( table->nelements > 0 ) {
	    fprintf( stderr, "\n\n" );
	}
    }

    if ( contour->cross_vertices )
    {
	hashtable_t	*table = contour->cross_vertices;

	if ( table->nelements > 0 ) {
	    fprintf( stderr, "  c:" );
	}
	for ( i = 0; i < table->size; i++ )
	{
	    hashtable_elt_t	*elt = table->elements[ i ];
	    
	    while ( elt )
	    {
		tess_vertex_t	*v = HASH_VERTEX( elt->ptr );
		fprintf( stderr, " %d", v->index );
	
		elt = elt->next;
	    }
	}

	if ( table->nelements > 0 ) {
	    fprintf( stderr, "\n\n" );
	}
    }
#endif
}
